
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  See the file COPYING for details.
 *
 */

#include <config.h> /* autostuff */

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimpwidgets/gimpwidgets.h>

#include "plugin-intl.h" /* i18n - gettext */

#include "types.h"
#include "dialog.h"
#include "preview.h"

typedef struct {
  Glace_ClientData* clientData;
  gboolean* ui_auto;
} AutoUpdateData;

static GtkWidget* ace_dialog_new(GtkSignalFunc       ok_callback, 
				 GtkSignalFunc       close_callback,
				 GimpPDBStatusType*  status);

static gboolean   dialog_setup  (Glace_ClientData* const clientData);

static void       update_preview_cb(GtkWidget*          unused_widget, 
				    Glace_ClientData* const clientData);

static void       redraw_preview_cb(GtkWidget*          unused_widget, 
				    Glace_ClientData* const clientData);

static void       cancel_preview_cb(GtkWidget*          unused_widget, 
				    GlaceGimpPreview* const preview);

static void     auto_update_preview(GtkObject* unused_widget,
				    AutoUpdateData* const update_data);

static void     auto_update_preview_color(GtkObject* unused_widget,
					  AutoUpdateData* const update_data);

static void       toggle_cb        (GtkWidget*       toggle, 
				    gboolean*        value);

static void       auto_toggle_cb        (GtkWidget*       toggle, 
				    Glace_ClientData* const clientData);

static void       ok_callback   (GtkWidget*          widget, 
				 GtkWidget*          dlg);

static void       close_callback(GtkWidget*          unused_widget, 
				 GimpPDBStatusType*  status);

static void       omenu_callback(GtkWidget*          menuitem, 
				 Glace_ColorMethods* data);

static void       store_link     (GimpChainButton*   link_toggle, 
				  GimpPDBStatusType* status);

static void       strength_changed(GtkObject*    strength_scale, 
				   GtkObject*    bradj_scale);

static void       iter_changed(GtkAdjustment *adj, 
			       GtkProgressBar* progress);

static void       bradj_changed        (GtkObject* bradj_scale);
static void       adjust_double_changed(GtkAdjustment* adj, gdouble* variable);
static void       adjust_uint_changed  (GtkAdjustment* adj, guint* variable);

#define SPACING 4
#define HBOX_SPACING SPACING
#define DIALOG_BORDER_WIDTH 0
#define OUTER_BORDER_WIDTH 6
#define ACTION_AREA_BORDER_WIDTH OUTER_BORDER_WIDTH
#define TOP_AREA_BORDER_WIDTH OUTER_BORDER_WIDTH
#define TABLE_ROW_SPACING SPACING
#define TABLE_COL_SPACING SPACING

#define SPIN_ATTOPT_X GTK_EXPAND | GTK_FILL
#define SPIN_ATTOPT_Y 0

#define SCALE_ATTOPT_X GTK_EXPAND | GTK_FILL
#define SCALE_ATTOPT_Y 0

#define LABEL_ATTOPT 0

#define EXPAND (TRUE)
#define FILL (TRUE)
#define RESIZE (TRUE)
#define SHRINK (TRUE)

#define SCALE_WIDTH 80

gboolean
ace_dialog(const GimpDrawableID drawable_id, 
	   AceValues* const acevals,
	   GimpPDBStatusType* const status,
	   Glace_ClientData* const clientData)
{
  GlaceGimpPreview* ace_preview;
  
  GtkObject *iter_spin_adj, *size_spin_adj, *smooth_spin_adj;
  
  GtkWidget *dlg, *pane, *vbox, *hbox, *table, *frame, *align;
  GtkWidget *label, *hseparator;
  GtkObject *bradj_scale, *strength_scale;
  GtkWidget *link_toggle, *button, *toggle;
  GtkWidget *iter_spin, *size_spin, *smooth_spin;
  GtkWidget *menuitem, *color_menu, *color_omenu;
  
  GtkTooltips *tooltips;

  AutoUpdateData* const update_data = g_new0(AutoUpdateData,1);

#ifdef SHOW_THRESHOLD_CONTROL
  GtkWidget *foobarum_spin;
  GtkObject *foobarum_spin_adj;
#endif
#ifdef DIAGNOSTIC
  int i; 
  GtkWidget *preview;
#endif

  gboolean run = FALSE;

  dlg = ace_dialog_new((GtkSignalFunc) ok_callback,
		       (GtkSignalFunc) close_callback,
		       status);

  g_idle_add((GSourceFunc)dialog_setup,clientData);
  /* g_signal_connect(G_OBJECT(dlg), "show",
     G_CALLBACK(dialog_setup), clientData); */
		       
  tooltips=gtk_tooltips_new();
  
  pane=gtk_hpaned_new();
  
  /* Preview */

  update_data->clientData = clientData;
  update_data->ui_auto = &(acevals->ui_auto);

  clientData->wData->preview = ace_preview = 
	  ace_preview_new(clientData->wData);

  g_signal_connect (ace_preview->preview, "invalidated",
                    G_CALLBACK(redraw_preview_cb), clientData);

  
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(ace_preview->preview), 
		     EXPAND, FILL, 0);
  
  hbox = gtk_hbox_new(TRUE, SPACING);

  button=gtk_button_new_with_label(_("Cancel"));

  g_signal_connect(G_OBJECT(button),"clicked",
		     G_CALLBACK(cancel_preview_cb), ace_preview);

  gtk_box_pack_end(GTK_BOX(hbox), button, !EXPAND, !FILL, 0);

  toggle = gtk_check_button_new_with_label(_("Auto"));
  button = gtk_button_new_with_label(_("Update"));

  g_signal_connect(G_OBJECT(button),"clicked",
		     G_CALLBACK(update_preview_cb), clientData);
  
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle),acevals->ui_auto);

  g_signal_connect(G_OBJECT(toggle),"toggled",
		     G_CALLBACK(toggle_cb),
		     &(acevals->ui_auto));

  g_signal_connect(G_OBJECT(toggle),"toggled",
		     G_CALLBACK(auto_toggle_cb),
		     clientData);

  gtk_box_pack_end(GTK_BOX(hbox), button, !EXPAND, !FILL, 0);
  gtk_box_pack_end(GTK_BOX(hbox), toggle, !EXPAND, !FILL, 0);

  align = gtk_alignment_new(1,0.5,0,0);
  gtk_container_add(GTK_CONTAINER(align), hbox);
  gtk_box_pack_end(GTK_BOX(vbox), align, !EXPAND, !FILL, SPACING);
  
  gtk_paned_pack1(GTK_PANED(pane), vbox, RESIZE, !SHRINK);
  
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dlg)->vbox), pane,
		      EXPAND, FILL, 0);
  
  gtk_box_pack_end(GTK_BOX(vbox), GTK_WIDGET(ace_preview->pbar), 
		   !EXPAND, !FILL, 0);

#ifdef DIAGNOSTIC
  hbox = gtk_hbox_new(FALSE,0);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dlg)->vbox), hbox,
			EXPAND, FILL, 0);
  for(i=0;i<2;i++) {
    preview = gtk_preview_new(GTK_PREVIEW_GRAYSCALE);
    gtk_preview_size(GTK_PREVIEW(preview),sel_width,sel_height);
    scroll = gtk_scrolled_window_new(NULL,NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll),
					  GTK_WIDGET(preview));
    gtk_widget_set_usize(scroll,MIN(PREVIEW_WIDTH, sel_width),
			 MIN(PREVIEW_HEIGHT, sel_height));

    gtk_container_add(GTK_CONTAINER(hbox),scroll);

    clientData->wData->diag[i] = GTK_PREVIEW(preview);
  }
#endif

  /* Parameter settings */
  frame = gtk_frame_new(_("Parameter Settings"));

  gtk_paned_pack2(GTK_PANED(pane), frame, !RESIZE, !SHRINK);  

  vbox = gtk_vbox_new (FALSE, SPACING);
	
  gtk_container_add(GTK_CONTAINER(frame),vbox);

  gtk_container_set_border_width (GTK_CONTAINER (vbox), 
			      TOP_AREA_BORDER_WIDTH);
  
  /* Table which holds sliders */
  table = gtk_table_new (2, 3, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, !EXPAND, FILL, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
  gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);

  strength_scale = gimp_scale_entry_new (GTK_TABLE (table), 0, 0, 
                                         _("Strength:"), SCALE_WIDTH, 0,
                                         acevals->strength, 0, 1,
                                         0.01, 0.1, 2,
                                         TRUE, 0, 0,
                                         NULL, NULL);

  bradj_scale = gimp_scale_entry_new (GTK_TABLE (table), 0, 1,
                                      _("Brightness Adjust:"), SCALE_WIDTH, 0,
                                      acevals->bradj, 0, 1,
                                      0.01, 0.1, 2,
                                      TRUE, 0, 0,
                                      NULL, NULL);
  
  /* Link toggle */  

  link_toggle = gimp_chain_button_new(GIMP_CHAIN_RIGHT);

  gimp_chain_button_set_active(GIMP_CHAIN_BUTTON(link_toggle),
			       acevals->ui_link);

  g_object_set_data(G_OBJECT(link_toggle),
		      "acevals.ui_link", &(acevals->ui_link));
  g_signal_connect(G_OBJECT(link_toggle), "destroy",
		     G_CALLBACK(store_link), status);

  gtk_table_attach (GTK_TABLE (table), link_toggle, 3, 4, 0, 2,
		    LABEL_ATTOPT, LABEL_ATTOPT, 0, 0);

  /* Callbacks for above adjustments */
  g_object_set_data(G_OBJECT(strength_scale), "user_data",
			   &(acevals->strength));
  g_signal_connect (G_OBJECT (strength_scale), "value_changed",
		      G_CALLBACK (strength_changed), 
		      bradj_scale);

  g_object_set_data(G_OBJECT(bradj_scale), "user_data",
			   &(acevals->bradj));

  g_object_set_data(G_OBJECT(bradj_scale),
		      "link", link_toggle);

  g_signal_connect (G_OBJECT (bradj_scale), "value_changed",
		      G_CALLBACK (bradj_changed), 
		      NULL);

  /* auto-update callbacks */
	
  g_signal_connect (G_OBJECT (strength_scale), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);

  g_signal_connect (G_OBJECT (bradj_scale), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);

  /* Seperator */
  hseparator = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (vbox), hseparator, !EXPAND, !FILL, 0);

  /* Table which holds spinbuttons */
  /* (iterations, size, threshold) */

  table = gtk_table_new (2, 4, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox), table, !EXPAND, !FILL, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table), SPACING);
  gtk_table_set_col_spacings (GTK_TABLE (table), SPACING);

  /* Iteration spin */
  iter_spin_adj = gtk_adjustment_new (acevals->iterations, 
				      1, 40, 1, 5, 0);
  iter_spin = gtk_spin_button_new (GTK_ADJUSTMENT (iter_spin_adj), 
				   1, 0);
  gtk_widget_set_name (iter_spin, "iteration_spin");
  gtk_table_attach (GTK_TABLE (table), iter_spin, 0, 1, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);
  
  g_object_set_data(G_OBJECT(iter_spin_adj), "user_data",
			   &(acevals->iterations));
  
  g_signal_connect (G_OBJECT (iter_spin_adj), "value_changed",
		      G_CALLBACK (adjust_uint_changed), 
		      &(acevals->iterations));
  
  g_signal_connect (G_OBJECT (iter_spin_adj), "value_changed",
		      G_CALLBACK (iter_changed), 
		      ace_preview->pbar);

  g_signal_connect (G_OBJECT (iter_spin_adj), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);

  /* Iteration label */
  label = gtk_label_new (_("Iterations:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
  
  /* size spin */
  size_spin_adj = gtk_adjustment_new (acevals->wsize, 
				      4, 128, 1, 10, 0);
  size_spin = gtk_spin_button_new (GTK_ADJUSTMENT (size_spin_adj), 1, 0);
  gtk_widget_set_name (size_spin, "size_spin");
  gtk_table_attach (GTK_TABLE (table), size_spin, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 
		    0, 0);
  g_object_set_data(G_OBJECT(size_spin_adj), "user_data",
			   &(acevals->wsize));
  g_signal_connect (G_OBJECT (size_spin_adj), "value_changed",
		      G_CALLBACK (adjust_uint_changed), 
		      &(acevals->wsize));

  g_signal_connect (G_OBJECT (size_spin_adj), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);
  
  /* size label */
  label = gtk_label_new (_("Detail Size:"));
  gtk_table_attach (GTK_TABLE (table), label, 1, 2, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
  
  /* smoothing spin */
  smooth_spin_adj = gtk_adjustment_new (acevals->smoothing,
					0, 100, 5, 10, 0);
  smooth_spin = gtk_spin_button_new (GTK_ADJUSTMENT (smooth_spin_adj),
				     1, 0);
  gtk_widget_set_name (smooth_spin, "smooth_spin");
  gtk_table_attach (GTK_TABLE (table), smooth_spin, 2, 3, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);
  g_object_set_data(G_OBJECT(smooth_spin_adj), "user_data",
			   &(acevals->smoothing));
  g_signal_connect (G_OBJECT (smooth_spin_adj), "value_changed",
		      G_CALLBACK (adjust_double_changed), 
		      &(acevals->smoothing));

  g_signal_connect (G_OBJECT (smooth_spin_adj), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);
  
  /* smoothing label */
  label = gtk_label_new (_("Smoothing:"));
  gtk_table_attach (GTK_TABLE (table), label, 2, 3, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
 
#ifdef SHOW_THRESHOLD_CONTROL 
  /* foobarum spin */
  foobarum_spin_adj = gtk_adjustment_new (acevals->coefftol,
					  0, 1, .0005, .002, 0);
  foobarum_spin = gtk_spin_button_new(GTK_ADJUSTMENT(foobarum_spin_adj),
				      1, 3);
  gtk_widget_set_name (foobarum_spin, "foobarum_spin");
  gtk_table_attach (GTK_TABLE (table), foobarum_spin, 3, 4, 1, 2,
		    SPIN_ATTOPT_X, SPIN_ATTOPT_Y,
		    0, 0);
  g_object_set_data(GTK_OBJECT(foobarum_spin_adj), "user_data",
			   &(acevals->coefftol));
  g_signal_connect (G_OBJECT (foobarum_spin_adj), "value_changed",
		      G_CALLBACK (adjust_double_changed), 
		      &(acevals->coefftol));

  g_signal_connect (G_OBJECT (foobarum_spin_adj), "value_changed",
		      G_CALLBACK (auto_update_preview), 
		      update_data);
 
  /* foobarum label */
  label = gtk_label_new (_("Threshold:"));
  gtk_table_attach (GTK_TABLE (table), label, 3, 4, 0, 1,
		    LABEL_ATTOPT, LABEL_ATTOPT,
		    0, 0);
#endif  

  /** Color method **/
  frame=gtk_frame_new(_("Color Method"));
  
  color_menu=gtk_menu_new();
  
  menuitem = gtk_menu_item_new_with_label(_("add white"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_DESATURATE));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("+correctdark"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_DESATURATE_DARK));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("limit"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_LIMIT));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("+correctdark"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_LIMIT_DARK));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("gray only"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_GRAY));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("0.6.2 white"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_SAT_ALEX));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("+correctdark"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_SAT_ALEX_DARK));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("+correctlight"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_SAT_MARTIN));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  menuitem = gtk_menu_item_new_with_label(_("correctlight+dark"));
  gtk_menu_shell_append(GTK_MENU_SHELL(color_menu), menuitem);
  g_object_set_data(G_OBJECT(menuitem), "user_data",
			   GUINT_TO_POINTER(GLACE_COLOR_SAT_MARTIN_DARK));
  g_signal_connect(G_OBJECT(menuitem), "activate",
		     G_CALLBACK(omenu_callback),
		     &(acevals->color_method));
  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(omenu_callback),
		     &(clientData->cfgInfoPtr->colorMethod));

  g_signal_connect(G_OBJECT(menuitem),"activate",
		     G_CALLBACK(auto_update_preview_color),
		     update_data);
  
  gtk_widget_show_all(color_menu);
  
  color_omenu = gtk_option_menu_new();
  gtk_option_menu_set_menu(GTK_OPTION_MENU(color_omenu),
			   color_menu);
  
  switch(acevals->color_method) {
  case GLACE_COLOR_DESATURATE:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 0);
	  break;
  case GLACE_COLOR_DESATURATE_DARK:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 1);
	  break;
  case GLACE_COLOR_LIMIT:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 2);
	  break;
  case GLACE_COLOR_LIMIT_DARK:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 3);
	  break;
  case GLACE_COLOR_GRAY:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 4);
	  break;
  case GLACE_COLOR_SAT_ALEX:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 5);
	  break;
  case GLACE_COLOR_SAT_ALEX_DARK:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 6);
	  break;
  case GLACE_COLOR_SAT_MARTIN:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 7);
	  break;
  case GLACE_COLOR_SAT_MARTIN_DARK:
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 8);
	  break;
  case GLACE_COLOR_LAST:
  default:
	  /* Handle bad values with only a little noise. */
	  g_assert_not_reached();
	  gtk_option_menu_set_history(GTK_OPTION_MENU(color_omenu), 1);
  }
  
  gtk_container_add(GTK_CONTAINER(frame), color_omenu);
  gtk_box_pack_start (GTK_BOX(vbox), frame, FALSE, TRUE, 0);

  if(!gimp_drawable_is_rgb(drawable_id)) 
    gtk_widget_set_sensitive(frame, FALSE);
  
  /* done */
  
  gtk_widget_show_all(dlg);
  
  run = (gimp_dialog_run (GIMP_DIALOG (dlg)) == GTK_RESPONSE_OK);

  return run;
}

static GtkWidget*
ace_dialog_new(GtkSignalFunc ok_callback, 
	       GtkSignalFunc close_callback,
	       GimpPDBStatusType *status)
{
  GtkWidget *dlg;
  gchar *wintitle;
  
  gimp_ui_init(PLUGIN_NAME, TRUE);
  
  wintitle = g_strdup_printf(_("Adaptive Contrast Enhancement v %s"), 
			     PLUGIN_VERSION);	
  
  dlg = gimp_dialog_new (wintitle,
			 PLUGIN_NAME,
                         NULL,
                         0,
			 gimp_standard_help_func, 
                         "gimp-ace",
			 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			 GTK_STOCK_OK,     GTK_RESPONSE_OK,
                         
			 NULL);
  g_free(wintitle);
 
#if 0  
  /* FIXME: is gimp_help_int obsolete? */
  gimp_help_init ();
#endif  

  return dlg;
}

/* The idea behind this function is it does some time-consuming set-up
 * operations, *after* there's been a chance for the dialog to show
 * itself.                                                             */
static gboolean dialog_setup(Glace_ClientData* const clientData) 
{ 
  static guint8 stage=0;

  switch(stage) 
    {
    case 0:
      gimp_ace_init_2(clientData);
      break;
    case 1:
      clientData->wData->target=GLACE_GIMP_PREVIEW;
      break;
    case 2:
      ace_preview_update(clientData);
      break;
    }
  stage++; 
  return (stage < 3);
}

/****  Preview-area callbacks. */

static void
update_preview_cb(GtkWidget* G_GNUC_UNUSED unused_widget,
		  Glace_ClientData* const clientData)
{
  ace_preview_update(clientData);
}

static void
redraw_preview_cb(GtkWidget* G_GNUC_UNUSED unused_widget,
		  Glace_ClientData* const clientData)
{
  Glace_Redraw_Preview(clientData->wData,clientData->imgArraysPtr); 
}

static void
cancel_preview_cb(GtkWidget* G_GNUC_UNUSED unused_widget,
		  GlaceGimpPreview* const preview)
{
  ace_preview_cancel(preview);
}

static void
toggle_cb (GtkWidget *toggle, gboolean* value)
{
  *value=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle));
}

static void
auto_toggle_cb (GtkWidget *toggle, Glace_ClientData* const clientData)
{
  if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle)) &&
     ace_preview_get_state(clientData->wData->preview) != PREVIEW_WORKING)
      ace_preview_update(clientData);
}

static void
auto_update_preview(GtkObject* G_GNUC_UNUSED unused_widget,
		    AutoUpdateData* const update_data)
{
  if(*(update_data->ui_auto)) ace_preview_update(update_data->clientData);
}

/* Changing the colour method is a special case of updating the preview,
   as we don't have to recompute the whole thing, just re-do the last bit...*/
static void
auto_update_preview_color(GtkObject* G_GNUC_UNUSED unused_widget,
			  AutoUpdateData* const update_data)
{
  if(*(update_data->ui_auto) && 
     ace_preview_get_state(update_data->clientData->wData->preview) != 
     PREVIEW_WORKING)
    {
      update_data->clientData->wData->target=GLACE_GIMP_PREVIEW;
      GlaceGimp_Output(update_data->clientData->wData,
		   update_data->clientData->cfgInfoPtr,
		   update_data->clientData->imgArraysPtr);
    }
}

/****** Parameter setting callbacks. *******/

static void
iter_changed(GtkAdjustment *adj, GtkProgressBar* pbar)
{
  static guint last_iter=0;
  
  last_iter = (guint)adj->value;
}

static void 
adjust_uint_changed (GtkAdjustment *adj, guint* variable)
{
  *variable=(guint)(adj->value);
}

static void 
adjust_double_changed (GtkAdjustment *adj, gdouble* variable)
{
  *variable=(gdouble)(adj->value);
}

static void 
strength_changed (GtkObject *strength_scale, GtkObject *bradj_scale)
{
  GimpChainButton *link_toggle = NULL;
  gdouble *strength;
  strength=g_object_get_data (G_OBJECT(strength_scale), "user_data");
  *strength=(gdouble)(GIMP_SCALE_ENTRY_SCALE_ADJ(strength_scale)->value);

  link_toggle=
    GIMP_CHAIN_BUTTON(g_object_get_data(G_OBJECT(bradj_scale),
					  "link"));
  
  if(gimp_chain_button_get_active(link_toggle))
    gtk_adjustment_set_value (GIMP_SCALE_ENTRY_SCALE_ADJ(bradj_scale), 
                              *strength);
}

static void 
bradj_changed (GtkObject *bradj_scale)
{
  gdouble *bradj_p=NULL;
  gdouble bradj=(gdouble)(GIMP_SCALE_ENTRY_SCALE_ADJ(bradj_scale)->value);
  
  bradj_p = g_object_get_data (G_OBJECT(bradj_scale), "user_data");
  *bradj_p = bradj;
}

static void
omenu_callback (GtkWidget *menuitem, Glace_ColorMethods* data)
{
  *data = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(menuitem), "user_data"));
}

static void
store_link(GimpChainButton* link_toggle, GimpPDBStatusType *status)
{
  if(*status != GIMP_PDB_CANCEL)
    {
      gboolean* link;
      link = (gboolean*)g_object_get_data(G_OBJECT(link_toggle),
					    "acevals.ui_link");
      *link=gimp_chain_button_get_active(link_toggle);
    }
}

/***********************************
 * Generic action-area callbacks.  */

static void 
ok_callback  (GtkWidget G_GNUC_UNUSED *unused_widget, 
	      GtkWidget* dlg)
{
  gtk_widget_destroy(dlg);
  gtk_main_quit();
}

static void 
close_callback  (GtkWidget G_GNUC_UNUSED *unused_widget, 
		 GimpPDBStatusType* status)
{
  *status=GIMP_PDB_CANCEL;
  gimp_quit();
}

#ifdef DIAGNOSTICS
/*************************************/
/* Diagnostic previews               */

/* Returns a Scrolled Window */
static GtkWidget*
diagnostic_preview_new(GtkPreview** preview_ptr)
{
  GtkWidget* preview = gtk_preview_new(GTK_PREVIEW_GRAYSCALE);
  
}

#endif

/* For emacs:
Local Variables:
mode:c
c-file-style:"gnu"
End:
*/
