/* $Id: gimp_ace.c,v 1.11.2.4 2000/08/30 09:07:00 acapnotic Exp $ */
/* Adaptive Contrast Enhancement plug-in for the GIMP (v1.2)
 * (A windowed histogram equalization using the Fourier-series based
 * fast algorithm.)
 * Based on J. Alex Stark's <stark@niss.org> GLACE system. 
 * GIMP Plug-in by Kevin Turner <kevint@poboxes.com> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.  See the file COPYING for details.
 *
 */

/* :HEADERS: 
 */

#include <stdlib.h>
#include <config.h>		/* autostuff */

#include "plugin-intl.h"

#include "preview.h"
#include "gimp_ace.h"		/* defines the AceValues struct */
#include "dialog.h"             /* because we call the dialog. */
#include "color.h"		/* This has the RGB -> Y conversion factors.
				   for the SeperateChannels */
#include "glaceInt.h"           /* GlaceGimp_Output uses ACC_BITS, OUT_BITS */
/* :*** Prototypes ***: 
 */

static void query (void);
static void run (const gchar           *name,
                 gint                   nparams,
                 const GimpParam       *param,
                 gint                  *nreturn_vals, 
                 GimpParam            **return_vals);

static Glace_ClientData* gimp_ace_init(GimpDrawableID drawable_id);

static void GlaceGimp_ReadImgArrays (Glace_ClientData* const clientData);

static int SeperateChannels (guchar inbuf[], const gint buflen,
			     const gboolean TypeRGB, const gboolean HasAlpha,
			     guchar ** Graybuf);

static gboolean GlaceGimp_ProcessIter(Glace_ClientData* const clientData);

#define CUBE(a)((a)*(a)*(a))

inline static int Ymax(float* xp, 
		       float* yp, 
		       int*   Yp, 
		       const gboolean add_white);

static GimpPDBStatusType* status;

#ifdef DIAGNOSTIC
static void fill_diagnostic(GtkPreview* const preview, 
			    const PreviewPixel* const data,
			    const guint width, 
			    const guint height);
#endif

#define IS_WITHIN(X,L,H) (((X) >= L) && ((X) <= H))

/* :*** Definitions ***:
 */

GimpPlugInInfo PLUG_IN_INFO = {
  NULL,				/* init_proc */
  NULL,				/* quit_proc */
  query,			/* query_proc */
  run,				/* run_proc */
};

static AceValues acevals = {
  1.0,				/* Strength (full) */
  0.0,				/* brightness adjust (none) */
  0.0,				/* foobarm (none) */
  0.0,				/* smoothing (none) */
  10,				/* iterations */
  25,				/* window size */
  GLACE_COLOR_DESATURATE,	/* Color method */
  TRUE,				/* link strength and brightness adj in dlg. */
  FALSE                         /* Automatically update preview */
};

static GimpParamDef ace_pdb_params[] = {
  /* Required params. */
  {GIMP_PDB_INT32, "run_mode",       "GimpRunModeType"},
  {GIMP_PDB_IMAGE, "image_id",       "(unused)"},
  {GIMP_PDB_DRAWABLE, "drawable_id", "Drawable to affect"},
  /* Our params. */
  {GIMP_PDB_FLOAT, "strength",   "[0,1] Magnitude of effect"},
  {GIMP_PDB_FLOAT, "bradj",      "[0,1] Magnitude of brightness adjust"},
  {GIMP_PDB_INT32, "iterations", "[1,] No. of iterations"},
  {GIMP_PDB_FLOAT, "fthreshold", "[FIXME;0 if unused] FOOBARUM threshold"},
  {GIMP_PDB_INT32, "size",       "[FIXME,] Detail size"},
  {GIMP_PDB_FLOAT, "smoothing",  "[0,] Smoothing"},
  {GIMP_PDB_INT32, "colormethod","\
  0=GLACE_COLOR_DESATURATE, \
  1=GLACE_COLOR_DESATURATE_DARK, \
  2=GLACE_COLOR_LIMIT, \
  3=GLACE_COLOR_LIMIT_DARK, \
  4=GLACE_COLOR_GRAY, \
  5=GLACE_COLOR_SAT_ALEX, \
  6=GLACE_COLOR_SAT_ALEX_DARK"}
};
static const int ace_pdb_nparams = sizeof(ace_pdb_params) / sizeof(GimpParamDef);

MAIN ()

static void 
query (void)
{
  gchar *help_path;
  gchar *help_uri;
  
  const int ace_pdb_nretvals=0;
  GimpParamDef* ace_pdb_retvals=NULL;

  gimp_plugin_domain_register (PLUGIN_NAME, LOCALEDIR);

  
  help_path = g_build_filename (DATADIR, "help", NULL);
  help_uri = g_filename_to_uri (help_path, NULL, NULL);
  g_free (help_path);

  gimp_plugin_help_register ("http://developer.gimp.org/gimp-ace/help",
                             help_uri);

  g_free (help_uri);

  gimp_install_procedure ("plug_in_ace",
			  _("Adaptive Contrast Enhancement"),
			  _("FIXME: Write ACE help."),
			  "Alex Stark, Kevin Turner,",
			  "Alex Stark, Kevin Turner,",
			  "1998-2000",
			  "<Image>/Filters/Colors/Adaptive Contrast...",
			  "GRAY*, RGB*", 
			  GIMP_PLUGIN, 
			  ace_pdb_nparams, 
			  ace_pdb_nretvals,
			  ace_pdb_params, 
			  ace_pdb_retvals);
}

static void run (const gchar           *name,
                 gint                   nparams,
                 const GimpParam       *param,
                 gint                  *nreturn_vals, 
                 GimpParam            **return_vals)
{
  static GimpParam retvals[1];
  const GimpRunMode run_mode = param[0].data.d_int32;
  const GimpDrawableID drawable = param[2].data.d_drawable;
  Glace_ClientData *clientData;
  gint i;
  gboolean dialog_ret;

  *nreturn_vals = 1;
  retvals[0].type = GIMP_PDB_STATUS;  
  status   = (GimpPDBStatusType*) &retvals[0].data.d_status;
  *return_vals = retvals;

  switch (run_mode)
    {
    case GIMP_RUN_NONINTERACTIVE:
      *status = GIMP_PDB_CALLING_ERROR;
     
      /* Check number of parameters. */
      if(nparams != ace_pdb_nparams) 
	{
	  g_warning("nparams is %d, should be %d " 
		       "in run() with GIMP_RUN_NONINTERACTIVE\n",
		       nparams, ace_pdb_nparams);
	  return;
	}
      
      /* Check parameter types. */
      for(i=0;i<ace_pdb_nparams;i++) 
	{
	  if(param[i].type != ace_pdb_params[i].type) 
	    {
	      g_warning("parameter %d has type %d, should be %d "
			   "in run() with GIMP_RUN_NONINTERACTIVE\n",
			   i, param[i].type,ace_pdb_params[i].type);
	      return;
	    }
	}

      /* Bounds check parameters. */
      i=3; 
      if(!(IS_WITHIN(param[i].data.d_float,0,1)))
	{
	  g_warning("parameter %d (strength) has value %f, "
		    "should be in [0,1] "
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_float);
	  return;
	}
      i++;
      if(!(IS_WITHIN(param[i].data.d_float,0,1)))
	{
	  g_warning("parameter %d (bradj) has value %f, "
		    "should be in [0,1] "
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_float);
	  return;
	}
      i++;
      if(!(param[i].data.d_int32 >= 1))
	{
	  g_warning("parameter %d (iterations) has value %d, "
		       "should be in >= 1"
		       "in run() with GIMP_RUN_NONINTERACTIVE\n",
		       i, param[3].data.d_int32);
	  return;
	}
      i++;
      if(!(param[i].data.d_float >= 0))
	{
	  g_warning("parameter %d (fthreshold) has value %f, "
		    "should be >= 0 "
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_float);
	  return;
	}
      i++;
      if(!(param[i].data.d_int32 >= 1))
	{
	  g_warning("parameter %d (size) has value %d, "
		    "should be in >= 1"
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_int32);
	  return;
	}
      i++;
      if(!(param[i].data.d_float >= 0))
	{
	  g_warning("parameter %d (smoothing) has value %f, "
		    "should be >= 0 "
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_float);
	  return;
	}
      if(!(IS_WITHIN(param[i].data.d_int32,0,GLACE_COLOR_LAST-1)))
	{
	  g_warning("parameter %d (colormethod) has value %d, "
		    "should be in [0,%d] "
		    "in run() with GIMP_RUN_NONINTERACTIVE\n",
		    i, param[3].data.d_int32, GLACE_COLOR_LAST-1);
	  return;
	}
      
      /* Assign parameters */
      i=3;
      acevals.strength     = param[i++].data.d_float;
      acevals.bradj        = param[i++].data.d_float;
      acevals.iterations   = param[i++].data.d_int32;
      acevals.coefftol     = param[i++].data.d_float;
      acevals.wsize        = param[i++].data.d_int32;
      acevals.smoothing    = param[i++].data.d_float;
      acevals.color_method = param[i++].data.d_int32;

      /* status is no longer a calling error. */
      *status = GIMP_PDB_EXECUTION_ERROR;      
      break;

    case GIMP_RUN_INTERACTIVE:
      *status = GIMP_PDB_EXECUTION_ERROR;
      gimp_get_data ("plug_in_ace", &acevals);

      clientData = gimp_ace_init(drawable);

      dialog_ret = ace_dialog(drawable, &acevals, status, clientData);

      if (! dialog_ret)
        {
          *status = GIMP_PDB_CANCEL;
          return;
        }
      
      clientData->wData->target=GLACE_GIMP_PixelRgn;
      clientData->imgArraysPtr->pixelBytePad = 
	gimp_drawable_has_alpha (drawable) ? 1 : 0;

      gimp_progress_init (_("Adaptive Contrast Enhancement:"));
      gimp_ace_go(clientData);

      gimp_set_data ("plug_in_ace", &acevals, sizeof (AceValues));

      *status = GIMP_PDB_SUCCESS;
      return;
      break;
    case GIMP_RUN_WITH_LAST_VALS:
      *status = GIMP_PDB_EXECUTION_ERROR;
      gimp_get_data ("plug_in_ace", &acevals);
      break;
    }

  if(*status != GIMP_PDB_CANCEL)
    {
      clientData = gimp_ace_init(drawable);
      gimp_ace_init_2(clientData);
      clientData->wData->target=GLACE_GIMP_PixelRgn;
      clientData->imgArraysPtr->pixelBytePad = 
	gimp_drawable_has_alpha (drawable) ? 1 : 0;
      gimp_progress_init (_("Adaptive Contrast Enhancement:"));
      gimp_ace_go(clientData);

      /* FIXME: see about this... */
      *status = GIMP_PDB_SUCCESS;
    }
}

static void Glace_CfgPutValues(
    Glace_CfgInfo *cfgInfoPtr)
{
  GLACE_CFG(dimensionSrc) = GLACE_COMMANDLINE;
  GLACE_CFG(windBaseW) = 25;
  GLACE_CFG(windBaseH) = 25;
  GLACE_CFG(numTerms) = 10;
  GLACE_CFG(heType) = GLACE_FACTOR;
  GLACE_CFG(heFactor) = 0.0;
  GLACE_CFG(prescaleMode) = GLACE_COMMANDLINE;
  GLACE_CFG(prescaleVal) = 1.0;
  GLACE_CFG(gaussCwindWidth) = 0.0;
  GLACE_CFG(gaussCwind) = TRUE;
  GLACE_CFG(hammingCwind) = TRUE;
  GLACE_CFG(chrome) = GLACE_GG;
  GLACE_CFG(addbackMode) = GLACE_COMMANDLINE;
  GLACE_CFG(passthruFactor) = 1.0;
  GLACE_CFG(inoffsetMode) = GLACE_COMMANDLINE;
  GLACE_CFG(inoffsetVal) = 127.5;
  GLACE_CFG(coeffTol) = 0.0;
}

static Glace_ClientData*
gimp_ace_init(GimpDrawableID drawable_id)
{

  Glace_ClientData *clientData;

  /* I don't know what most of this does.
     It's copied out of pnmace.c */
  clientData = Glace_AllocClientData ();
  Glace_WInit (GLACE_CDATA (wData));

  Glace_CfgInit (GLACE_CDATA (cfgInfoPtr));

  Glace_CfgPutValues (GLACE_CDATA (cfgInfoPtr));

#ifdef DIAGNOSTIC
  clientData->cfgInfoPtr->verbose = TRUE;
#else /* no diagnostics */
  clientData->cfgInfoPtr->verbose = FALSE;
#endif
  
  clientData->wData->drawable_id = drawable_id;

  clientData->wData->drawable_ptr=
    gimp_drawable_get (drawable_id);

  clientData->wData->iteration = 0;
#ifdef DIAGNOSTIC
  clientData->wData->diag[0] = clientData->wData->diag[1] = NULL;
#endif /* DIAGNOSTIC */

    {
      gint x1=0,y1=0,x2=0,y2=0;
      gimp_drawable_mask_bounds (drawable_id, &x1, &y1, &x2, &y2);
      clientData->wData->drawable_h = x2 - x1;
      clientData->wData->drawable_w = y2 - y1;
      clientData->wData->has_alpha = gimp_drawable_has_alpha(drawable_id);
    }

  return clientData;
}

/* Stage 2 of the initilization is done after the dialog is shown, as
   it may take a little time. */
void 
gimp_ace_init_2(Glace_ClientData* clientData) 
{
  Glace_DefaultTmpImg (GLACE_CDATA (cfgInfoPtr), GLACE_CDATA (imgArraysPtr));
  GlaceGimp_ReadImgArrays (clientData);
}

void
gimp_ace_go(Glace_ClientData* clientData)
{
  Glace_CfgInfo* const cfgInfoPtr = clientData->cfgInfoPtr;
  GlaceGimpData* const wData = clientData->wData;

  cfgInfoPtr->heFactor        = 1.0 - acevals.strength;
  cfgInfoPtr->passthruFactor  = 1.0 - acevals.bradj;
  cfgInfoPtr->numTerms        = acevals.iterations;
  cfgInfoPtr->windBaseW       = acevals.wsize;
  cfgInfoPtr->windBaseH       = acevals.wsize;
  cfgInfoPtr->coeffTol        = acevals.coefftol;
  cfgInfoPtr->gaussCwindWidth = acevals.smoothing;
  cfgInfoPtr->colorMethod     = acevals.color_method;

  Glace_CfgAllocSeriesVectors (GLACE_CDATA (cfgInfoPtr));

  Glace_CfgBeginToHeseries (GLACE_CDATA (cfgInfoPtr));

  Glace_CfgHeseriesToAddback (GLACE_CDATA (cfgInfoPtr));

  Glace_CFuncGen (GLACE_CDATA (cfgInfoPtr));

  Glace_CfgAddbackToEnd (GLACE_CDATA (cfgInfoPtr));

  Glace_AllocTables (GLACE_CDATA (cfgInfoPtr), GLACE_CDATA (tableInfoPtr));

  Glace_AllocImgArrays (GLACE_CDATA (cfgInfoPtr), GLACE_CDATA (imgArraysPtr));

  clientData->wData->iteration = 0;

  switch(WRAPPER(target))
    {
    case GLACE_GIMP_PixelRgn:
      while(GlaceGimp_ProcessIter (clientData));
      break;
    case GLACE_GIMP_PREVIEW:
      clientData->wData->preview->idle_handle = 
          g_idle_add ((GSourceFunc)GlaceGimp_ProcessIter, 
		      clientData);
      break;
    } /* esac target */
}


/* Assumptions: Input is one byte-per-pixel grayscale. */
static void
GlaceGimp_ReadImgArrays (Glace_ClientData* clientData)
{
  GlaceGimpData* const wData = clientData->wData;
  Glace_CfgInfo* const cfgInfoPtr = clientData->cfgInfoPtr;
  Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;
  GimpDrawable* const drawable_ptr = wData->drawable_ptr;
  GimpDrawableID const drawable_id = wData->drawable_id;

  GimpPixel *gimp_buf = NULL;
  gint gimp_size;
  GimpPixelRgn src_rgn;
  gint x1, y1, x2, y2;		/* Bounds of the selection */

  /* Null all image pointers. */
  GLACE_IMG (inImgPtr) =
    GLACE_IMG (inImgCxPtr) =
    GLACE_IMG (inImgCyPtr) =
    GLACE_IMG (inImgCYMaxPtr) =
    GLACE_IMG (inImgRPtr) =
    GLACE_IMG (inImgGPtr) = 
    GLACE_IMG (inImgBPtr) = NULL;

  gimp_drawable_mask_bounds (drawable_id, &x1, &y1, &x2, &y2);

  WRAPPER(gimp_x0) = x1;
  WRAPPER(gimp_y0) = y1;
  GLACE_IMG(cols)  = x2 - x1;
  GLACE_IMG(rows)  = y2 - y1;

  gimp_pixel_rgn_init (&src_rgn,
		       WRAPPER (drawable_ptr),
		       WRAPPER (gimp_x0), WRAPPER (gimp_y0),
		       GLACE_IMG (cols), GLACE_IMG (rows), FALSE, FALSE);

  /* skip some pixels on output so we can mix the alpha back in. */
  GLACE_IMG (pixelBytePad) =
    gimp_drawable_has_alpha (drawable_id) ? 1 : 0;

  /* This refers to precision, not number of channels. */
  GLACE_IMG (pixelSize) = 1;

  gimp_size = GLACE_IMG (cols) * GLACE_IMG (rows) * drawable_ptr->bpp;

  WRAPPER(gimpImgPtr) = gimp_buf = g_new (GimpPixel, gimp_size);

  gimp_pixel_rgn_get_rect (&src_rgn, gimp_buf,
			   WRAPPER (gimp_x0), WRAPPER (gimp_y0),
			   GLACE_IMG (cols), GLACE_IMG (rows));

  imgArraysPtr->inImgSize =
    SeperateChannels (gimp_buf, gimp_size,
		      gimp_drawable_is_rgb (drawable_id), 
		      gimp_drawable_has_alpha (drawable_id), 
		      &imgArraysPtr->inImgPtr);	/* Greybuf (to process) */

  imgArraysPtr->refImgPtr = imgArraysPtr->inImgPtr;

  imgArraysPtr->refImgLBPtr = imgArraysPtr->refImgHBPtr =
    imgArraysPtr->inImgLBPtr = imgArraysPtr->inImgHBPtr =
    imgArraysPtr->inImgPtr;

  GLACE_CFG (chrome) = gimp_drawable_is_rgb (drawable_id)
    ? GLACE_CC : GLACE_GG;

  /* This is forced config stuff. */
  /*  if (GLACE_CFG(filtMethod) == GLACE_WINDOW) */
  GLACE_CFG (filtMethod) = GLACE_WINDOW;
  GLACE_IMG (ddHImgPtr) = GLACE_IMG (ddVImgPtr) = NULL;

  GLACE_CFG (doublep) = FALSE;

  {
    gint x1=0,y1=0,x2=0,y2=0;
    gimp_drawable_mask_bounds (drawable_id, &x1, &y1, &x2, &y2);

    GLACE_IMG(rowstride) = x2-x1;   
  }
}

static int
SeperateChannels (guchar inbuf[], const gint buflen,
		  const gboolean TypeRGB, const gboolean HasAlpha,
		  guchar ** Graybuf)
{
  int c = 0, ic = 0;
  guchar *Grayfoo;

  if (!TypeRGB && !HasAlpha) /* if GRAY */
    {
      *Graybuf = inbuf;
      return buflen;
    }

  if (!TypeRGB && HasAlpha) /* if GRAYA */
    {
      if ((buflen % 2) != 0) 
	{
	  g_assert_not_reached();
	  g_warning ("GRAYA with odd size %d?", buflen);
	  *status = GIMP_PDB_CALLING_ERROR;
	  gimp_quit();
	}
      if (*Graybuf == NULL)
	*Graybuf = g_new (guchar, buflen / 2);

      Grayfoo = *Graybuf;
      for(ic=0,c=0;ic<buflen;ic+=2,c++)
	{
	  Grayfoo[c] = inbuf[ic];
	}
      return buflen/2;
    } /* endif GRAYA */

  if (TypeRGB)
    {
      const int bpp = HasAlpha ? 4 : 3;
      if ((buflen % bpp) != 0)
	{
	  g_assert_not_reached();
	  g_warning("RGB: %d Alpha: %d.  I think bpp should be %d, but that"
		    "doesn't make sense with size %d, as R = %d",
		    TypeRGB, HasAlpha, bpp, buflen, buflen % bpp);
	  *status = GIMP_PDB_CALLING_ERROR;
	  gimp_quit();
	}

      if (*Graybuf == NULL)
	*Graybuf = g_new (guchar, buflen / bpp);

      Grayfoo = *Graybuf;

      for(ic=0,c=0;ic<buflen;ic+=bpp,c++)
	{
	  Grayfoo[c] = 
	    Y_r * inbuf[ic+0] + 
	    Y_g * inbuf[ic+1] + 
	    Y_b * inbuf[ic+2];
	}			/* whend */
      return buflen / bpp;
    }				/* endif rgb */

  g_assert_not_reached();
  *status=GIMP_PDB_EXECUTION_ERROR;
  gimp_quit();

  return -1;
}				/* SeperateChannels() */


/* Iterates through one pass of the Glace_Process loop.
 * Returns FALSE when done.
 * Suitable for use as a callback on an idle handler. */
static gboolean
GlaceGimp_ProcessIter(Glace_ClientData* const clientData)
{
  Glace_WData const wData = clientData->wData;
  Glace_CfgInfo* const cfgInfoPtr = clientData->cfgInfoPtr;
  Glace_ImgArrays* const imgArraysPtr = clientData->imgArraysPtr;
  Glace_TableInfo* const tableInfoPtr = clientData->tableInfoPtr;
  GlaceGimpPreview* const preview = wData->preview;

  if(wData->iteration == 0)
    {
      WRAPPER(errorValue)=GLACE_OK;
      Glace_WindChk (cfgInfoPtr, GLACE_IMG (cols), GLACE_IMG (rows));
      if(WRAPPER(errorValue)==GLACE_ERROR)
	{
	  /* FIXME: Can we handle this error by adjusting the window size
	     instead? */
	  gimp_message(_("A tiny image,\n"
			 "  too delicate for my hands.\n"
			 "Sorry, I can't do."));
	  /* FIXME: We quit without setting our return status to an error
	     code. */
	  gimp_quit();
	}

      Glace_InitAccIm (cfgInfoPtr, imgArraysPtr);
      
      wData->iteration = GLACE_CFG(firstTerm);
    }

  if(wData->iteration <= (unsigned) GLACE_CFG (numTerms))
    {
      Glace_SetForTerm (tableInfoPtr, wData->iteration, GLACE_COS);

#ifdef DIAGNOSTIC
      Glace_TermlyReport (cfgInfoPtr, tableInfoPtr);
      Glace_WMessage (GLACE_CFG (wData), ";\n");
#endif
      
      Glace_SetTables (cfgInfoPtr, tableInfoPtr);
      Glace_GenGen (cfgInfoPtr, tableInfoPtr, imgArraysPtr);
      
      Glace_SetForTerm (tableInfoPtr, wData->iteration, GLACE_SIN);
      
      Glace_SetTables (cfgInfoPtr, tableInfoPtr);
      Glace_GenGen (cfgInfoPtr, tableInfoPtr, imgArraysPtr);
            
      switch(WRAPPER(target))
	{
	case GLACE_GIMP_PixelRgn:
	  gimp_progress_update (0.10 + 0.90 * wData->iteration / 
				GLACE_CFG (numTerms));
	  break;
	case GLACE_GIMP_PREVIEW:
	  ace_preview_progress_update(preview, (gfloat) 
				      (PROG_INPUT_CHUNKS + wData->iteration) / 
				      (GLACE_CFG(numTerms) + PROG_OCHUNKS));
	  break;
	} /* esac target */
      wData->iteration++;
      return TRUE;
    }
  else /* Last iteration is done. */
    {
      GlaceGimp_Output (wData, cfgInfoPtr, imgArraysPtr);  
      switch(WRAPPER(target))
	{
	case GLACE_GIMP_PixelRgn:
	  break;
	case GLACE_GIMP_PREVIEW:
	  ace_preview_update_finished(preview);
	  break;
	} /* esac target */

      wData->iteration=0;
      return FALSE;
    } /* endif more iterations */
}

int
GlaceGimp_Output (Glace_WData wData,
		  Glace_CfgInfo *cfgInfoPtr,
		  Glace_ImgArrays *imgArraysPtr)
{
  Glace_MidGray temp_gray;
  const Glace_MidGray out_offset= 255 << (ACC_BITS - OUT_BITS);
  int r, c;
  Glace_MidGray *agP;
  const Glace_IOGray *inP;
  Glace_IOGray *ogP, tmph, tmpl;

  long temp_long;

  register int gstep;

  const gboolean grayout=(GLACE_CFG(chrome) != GLACE_CC);
  const gboolean doubleout=GLACE_CFG(doubleout);
  const gboolean doClip=GLACE_CFG(doClip);
  const Glace_ColorMethods colorMethod=GLACE_CFG(colorMethod);
  const gboolean correctDark=
    GLACE_CFG(colorMethod) ==GLACE_COLOR_SAT_MARTIN_DARK ||
    GLACE_CFG(colorMethod) ==GLACE_COLOR_DESATURATE_DARK ||
    GLACE_CFG(colorMethod) ==GLACE_COLOR_LIMIT_DARK ||
    GLACE_CFG(colorMethod) ==GLACE_COLOR_SAT_ALEX_DARK;
  
#ifdef DIAGNOSTIC
  guint count=0,count_gt1=0,count_gtmax=0,count_up=0,count_down=0;
  gulong total_plus_Y=0, total_minus_Y=0;
  PreviewPixel dpixel=255;
  static PreviewPixel* rowzeros=NULL;
  if(rowzeros==NULL) rowzeros=g_new0(PreviewPixel,GLACE_IMG(cols));
#endif

#ifdef THIS_MAKES_SENSE_TO_WHO
  /* 255 because that brings zero to the right (?) place and
     1 because that becomes 0.5, that is rounding by truncation. */
  if ((GLACE_CFG(chrome)==GLACE_CC) || (GLACE_CFG(doubleout)==TRUE))
    round_off=0;
  else
    round_off=1;

  out_offset = (255+round_off) << (ACC_BITS - OUT_BITS);
#endif /* makes sense? */
  
  if(doClip) g_warning("UNTESTED: Clipping field (doClip) has not been "
		       "tested in GlaceGimp_Output\n");

  if(doubleout && !grayout) 
    { 
      g_warning("GlaceGimp_Output doesn't even pretend to handle "
		"16-bit colour output.  Returning -1.\n");
      return -1;
    }

  if(doubleout) g_warning("UNTESTED: double-precision output has not been "
		       "tested in GlaceGimp_Output\n");			  

  /* We output row-by-row */
  Glace_WPutImgStart(wData,cfgInfoPtr,imgArraysPtr);

  agP = GLACE_IMG(outAccImgPtr);
  for (r = 0; r < GLACE_IMG(rows); r++)
    {      
      Glace_WPutImgRowStart(wData, cfgInfoPtr, imgArraysPtr, r);

#ifdef DIAGNOSTIC
	  if(WRAPPER(target)==GLACE_GIMP_PREVIEW)
	    {
	      gtk_preview_draw_row(WRAPPER(diag[0]),rowzeros, 
				   0,r,GLACE_IMG(cols));
	      gtk_preview_draw_row(WRAPPER(diag[1]),rowzeros, 
				   0,r,GLACE_IMG(cols));
	    }
#endif

      if (grayout)
	{
	  if(!doubleout)
	    {
	      gstep=GLACE_IMG(pixelBytePad);
	      if (!doClip)
		{
		  for (c=GLACE_IMG(cols), ogP = GLACE_IMG(putImgRowPtr);
		       c >0 ; c--, ogP++, ogP +=gstep, agP++)
		    {
		      temp_gray = (*agP + out_offset)
			>> (ACC_BITS - OUT_BITS + 1);
		      /*temp_gray = (Glace_MidGray) (((float) *agP)*tmpy +tmpz);*/
		      *ogP = temp_gray > 255 ? 255 : temp_gray < 0 ? 0 : temp_gray;
		    }
		}
#ifdef DIAGNOSTIC
	      else /* doClip - FIXME - untested */
		{
		  for (c = GLACE_IMG(cols), ogP = GLACE_IMG(putImgRowPtr);
		       c >0 ; c--, ogP++, ogP +=gstep, agP++)
		    {
		      temp_gray = (*agP + out_offset)
			>> (ACC_BITS - OUT_BITS + 1);
		      /*temp_gray = (Glace_MidGray) (((float) *agP)*tmpy +tmpz);*/
		      *ogP = temp_gray > 255 ? 255 : temp_gray < 0 ? 0 : 127;
		    } /* next pixel */
		} /* endif !doClip */
#endif
	    } 
	  else /* doubleout - FIXME - untested */
	    {
	      gstep = 3+GLACE_IMG(pixelBytePad);
	      for (c = GLACE_IMG(cols), ogP = GLACE_IMG(putImgRowPtr);
		     c >0; c--, ogP+= gstep,++agP )
		{
		  temp_long = (*agP + out_offset)
		    << (8-(ACC_BITS - OUT_BITS + 1));
		  temp_long = temp_long > 65535 ? 65535 : 
		    temp_long < 0 ? 0 : temp_long;
		  BYTE_SPLIT(temp_long, tmph, tmpl);
		  GLACE_ASSIGN(
			       *((Glace_Pixel *)ogP),
			       tmph,tmph,tmpl);
		} /* next pixel */
	    } /* endif doubleout */
	} /* endif grayout */
      else /* !grayout */
	{
	  const int instep = WRAPPER(drawable_ptr)->bpp;
	  gstep = 3 + GLACE_IMG(pixelBytePad);
	  for ( c = GLACE_IMG(cols),
		  ogP = GLACE_IMG(putImgRowPtr),
		  inP = WRAPPER(gimpImgPtr) + r * GLACE_IMG(cols) * instep;
		c >0 ;
		c--, ogP+=gstep, inP+=instep, ++agP )
	    {
	      gint vR=inP[0];
	      gint vG=inP[1];
	      gint vB=inP[2];
	      const gint beta0 = MAX(MAX(vR, vG), vB);
	      gint vY2 =  (*agP + out_offset) / (1<<(ACC_BITS - OUT_BITS + 1));
	      /* We do "a / (1 << b)" instead of "a >> b" because the behaviour
		 of "a >> b" is implementation dependant when a is negative.
		 We could cook up an autoconf test for this if we were really
		 worried about that extra division... */
	      if((colorMethod == GLACE_COLOR_GRAY) || (vR==vG && vR==vB) || 
		 (vR<(1/Y_r + 0.5) && vG < (1/Y_g + 0.5) && vB < (1/Y_b +0.5)))
		vR=vG=vB=CLAMP0255(vY2);
	      else
		switch(colorMethod)
		  {
		  case GLACE_COLOR_LIMIT:
		  case GLACE_COLOR_LIMIT_DARK:
		  case GLACE_COLOR_DESATURATE:
		  case GLACE_COLOR_DESATURATE_DARK:
		  {
		      gint c_changed = 0;
		      gint       vX = X_r * vR + X_g * vG + X_b * vB;
		      const gint vY1= Y_r * vR + Y_g * vG + Y_b * vB;
		      gint       vZ = Z_r * vR + Z_g * vG + Z_b * vB;
		      float vx,vy;
		      gint vY2orig = CLAMP0255(vY2);

		      if(vY2 > vY1)
		        {
		          vx = (float)vX /(vX+vY1+vZ);
			  vy = (float)vY1/(vX+vY1+vZ);
			  c_changed = Ymax(&vx,&vy,&vY2,
					(colorMethod !=
					GLACE_COLOR_LIMIT) &&
					(colorMethod != GLACE_COLOR_LIMIT_DARK));
		        }
		      else
		        {
			  if(vY2 < 0) vY2 = 0;
		        }

		      if(c_changed)
			{
			  vX = vY2 * vx/vy;
			  vZ = vY2 * (1.0 - vx - vy) / vy;

			} 
		      else 
			{
			  /* simplifies to: */
			  vX = vX * vY2 / vY1;
			  vZ = vZ * vY2 / vY1;
			}
		      
		      vR = CLAMP0255(R_x*vX + R_y*vY2 + R_z*vZ + 0.5);
		      vG = CLAMP0255(G_x*vX + G_y*vY2 + G_z*vZ + 0.5);
		      vB = CLAMP0255(B_x*vX + B_y*vY2 + B_z*vZ + 0.5);

		      if(correctDark && beta0 < 127)
			{
			  float correctdark = CUBE ((255 - beta0) / 255.0);
			  vR = vR * (1.0 - correctdark) + correctdark * vY2orig;
			  vG = vG * (1.0 - correctdark) + correctdark * vY2orig;
			  vB = vB * (1.0 - correctdark) + correctdark * vY2orig;
			}
		    }
		    break;
		  case GLACE_COLOR_SAT_ALEX:
		  case GLACE_COLOR_SAT_ALEX_DARK:
		  case GLACE_COLOR_SAT_MARTIN:
		  case GLACE_COLOR_SAT_MARTIN_DARK:
		    {
		      const double YMAP_ADD=1.0;
		      const Glace_IOGray mmaxval = GLACE_MAXMAXGRAY;
		      const gint32 vY1=Y_r * vR + Y_g * vG + Y_b * vB;
		      const int vR0=vR;
		      const int vG0=vG;
		      const int vB0=vB;		      
		      double lambda;
		      int alpha, beta;
		      
		      vY2 = CLAMP0255(vY2);
		      
		      
		      vR -= vY1;
		      vG -= vY1;
		      vB -= vY1;

		      beta = MAX(MAX(vR, vG), vB);
		      alpha = abs(MIN(MIN(vR, vG), vB));
		      
		      lambda = MIN( beta*vY2, alpha*(mmaxval-vY2) ) /
			(YMAP_ADD+MIN( beta*vY1, alpha*(mmaxval-vY1) ) );
		      
		      vR = vY2 + lambda* vR;
		      vG = vY2 + lambda* vG;
		      vB = vY2 + lambda* vB;
		      
		      if(colorMethod==GLACE_COLOR_SAT_MARTIN || 
			 colorMethod==GLACE_COLOR_SAT_MARTIN_DARK)
			{
			  float correctlight;
			  double vSum = (vR0 + vG0 + vB0)/(255.0*3);

                          correctlight = SQR(SQR(SQR (vSum)));

                          vR = vR * (1.0 - correctlight) + correctlight * vY2;
			  vG = vG * (1.0 - correctlight) + correctlight * vY2;
			  vB = vB * (1.0 - correctlight) + correctlight * vY2;
			}

		      if(correctDark && beta0 < 127)
			{
			  float correctdark = CUBE ((255 - beta0) / 255.0);
			  vR = vR * (1.0 - correctdark) + correctdark * vY2; 
			  vG = vG * (1.0 - correctdark) + correctdark * vY2;
			  vB = vB * (1.0 - correctdark) + correctdark * vY2;
			}
		    }
		    break;
		  case GLACE_COLOR_GRAY:
		    /* is handled before this switch */
		  case GLACE_COLOR_LAST:
		    /* just a placeholder */
		  default:
		    g_warning("bad color method %d",colorMethod);
		    g_assert_not_reached();
		} /* esac colorMethod */

	      GLACE_ASSIGN(*((Glace_Pixel *)ogP),
			   vR, vG, vB);
#ifdef DIAGNOSTIC
	      count++;
#endif
	    } /* next pixel */
	} /* endif !grayout */      
      Glace_WPutImgRowFinish(wData, cfgInfoPtr, imgArraysPtr, r);
    } /* next row */

#ifdef DIAGNOSTIC
  if(WRAPPER(target)==GLACE_GIMP_PREVIEW)
  {
    gtk_widget_draw(WRAPPER(diag[0]));
    gtk_widget_draw(WRAPPER(diag[1]));
  }

  g_print("Pixels :\t%d\n" 
	  "Y2 > Y1:\t%d\t(%4.1f%% of total)\n"
	  "Y2 > YMax:\t%d\t(%4.1f%% of Y2 > Y1)\n"
	  "Y2 > 1 :\t%d\t(%4.1f%% of Y2 > Y1)\n",
	  count,
	  count_up, 100 * (float)count_up / (float)count,
	  count_gtmax, 100 * (float)count_gtmax / (float)count_up,
	  count_gt1, 100 *(float)count_gt1 / (float)count_up);
  g_print("avg rise:\t% 5f\n" 
	  "avg drop:\t% 5f\n"
	  "overall :\t%+5f\n",
	  (float)total_plus_Y/(float)count_up/255.0,
	  (float)total_minus_Y/(float)count_down/255.0,
	  (float)((signed)total_plus_Y - (signed)total_minus_Y)/
	  (float)count/255.0);
#endif /* DIAGNOSTIC */
  
  Glace_WPutImgFinish(wData, cfgInfoPtr, imgArraysPtr);
  
  return Glace_WErrorValue( GLACE_IMG(wData) );
}

/* This makes more sense if you look at the graph.
   run 'gnuplot < doc/max-luminosity.gnuplot' for a visual. */
/* x and y are chromacity coordinates. 
   the return is a luminosity scaled 0-255 */
inline static int
Ymax(float* const xp, float* const yp, int* const Yp, 
     const gboolean add_white)
{
  const int Y = *Yp;
  const double y = *yp;
  const double x = *xp;
  gint Ymax=255;
  typedef enum { RED_ROOF, GREEN_ROOF, BLUE_ROOF } RoofColor;
  RoofColor shingle;

  if(y > y_White) /* above Cyan-White line */
    {
      if(x > x_White) /* right of White */
	{
	  /* below white-yellow line? */
	  if((y - y_White) < 
	     ((y_Yellow - y_White) * (x - x_White) / (x_Yellow - x_White)))
	    {
	      shingle = RED_ROOF;
	    } else {
	      shingle = GREEN_ROOF;
	    } /* end white-yellow-line */
	} else {
	  shingle = GREEN_ROOF;
	} /* endif left-of-white */
    }
  else  /* Below Cyan-White line */
    {
      if ((x - x_White) < /* left of White-Purple line */
	  ((y - y_White) * (x_Purple-x_White)/(y_Purple-y_White)))
	{
	  shingle = BLUE_ROOF;
	} else {
	  shingle = RED_ROOF;
	}
    } /* endif below cyan-white */

  switch (shingle)
    {
    case RED_ROOF:
      Ymax= 255.0 /(R_y + R_x*(x)/(y) + R_z * (1.0-x-y) / y);
      break;
    case GREEN_ROOF:
      Ymax= 255.0 /(G_y + G_x*(x)/(y) + G_z * (1.0-x-y) / y);
      break;
    case BLUE_ROOF:
      Ymax= 255.0 /(B_y + B_x*(x)/(y) + B_z * (1.0-x-y) / y);
      break;
    }

  if(Y <= Ymax)
    {
      /* nop, everything remains unchanged. */
      return FALSE;
    } 
  else if(!add_white) /* Y > Ymax */
    {
      /* x and y unchanged */
      *Yp=Ymax;
      return FALSE;
    }
  else 
    if (Y >= 255) /* add_white && Y > 100% */
    {
      *xp=x_White;
      *yp=y_White;
      *Yp=255;
      return (Y - Ymax);
    }
  else /* add_white && 255 > Y > Ymax */
    {
      double x2, y2=0;
      const double dxdy = (x_White - x) / (y_White - y);
      switch (shingle)
	{
	case RED_ROOF:
	  y2=(x * (R_x - R_z) + y * dxdy * (R_z-R_x) + R_z) /
	     (255.0/Y + dxdy * (R_z - R_x) - R_y + R_z);
	  break;
	case GREEN_ROOF:
	  y2=(x * (G_x - G_z) + y * dxdy * (G_z-G_x) + G_z) /
	     (255.0/Y + dxdy * (G_z - G_x) - G_y + G_z);
	  break;
	case BLUE_ROOF:
	  y2=(x * (B_x - B_z) + y * dxdy * (B_z-B_x) + B_z) /
	     (255.0/Y + dxdy * (B_z - B_x) - B_y + B_z);
	  break;
	}
      x2 = x + dxdy * (y2 - y);
  
      *xp = x2;
      *yp = y2;
      
      /* Y is unchanged. */
      return (Y - Ymax); /* chromacity did change */
    } /* endif addwhite */
}

#ifdef DIAGNOSTIC
static void
fill_diagnostic(GtkPreview* const preview, const PreviewPixel* data,
		const guint width, const guint height)
{
  guint yy;
  for(yy=0; yy < height; yy++) {
    gtk_preview_draw_row(preview,data,0,yy,width);
    data+=width;
  }
}
#endif

/* For emacs:
Local Variables:
mode:c
c-file-style:"gnu"
End:
*/
